@@ -95,7 +95,7 @@ static const char empty_str[] = "";
 				gfarm_funcname, mount_point, \
 				path_prefix, gfarm_path + path_offset, \
 				gfarm_error_string(gfarm_e)); \
-		} else if (ret_errno != ENOENT) { \
+		} else if (ret_errno != ENOENT && ret_errno != ENOTSUP) { \
 			gflog_info(msgNo, syslog_fmt, fuse_opname, \
 				gfarm_funcname, mount_point, \
 				path_prefix, gfarm_path + path_offset, \
@@ -529,8 +529,16 @@ gfarm2fs_getattr(const char *path, struct stat *stbuf)
 				     "gfarmize_path", path, e);
 		return (-gfarm_error_to_errno(e));
 	}
+
+	/*
+	 * gfarm2fs_getattr and gfarm2fs_release may be called simultaneously
+	 * after write-close.
+	 * If there is no this lock, the filesize will be wrong.
+	 */
+	gfarm2fs_open_file_table_rdlock();
 	e = gfs_lstat_cached(gfarmized.path, &st);
 	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_open_file_table_unlock();
 		if (IS_SUBDIR(gfarmized.path) &&
 		    strcmp(gfarmized.path + gfarm2fs_subdir_len, "/" GFARM_DIR)
 		    == 0) {
@@ -555,17 +563,16 @@ gfarm2fs_getattr(const char *path, struct stat *stbuf)
 		free_gfarmized_path(&gfarmized);
 		return (-gfarm_error_to_errno(e));
 	}
-	gfarm2fs_open_file_table_lock();
-	if ((fp = gfarm2fs_open_file_lookup(st.st_ino)) != NULL) {
+	if ((fp = gfarm2fs_open_file_lookup_unlocked(st.st_ino)) != NULL) {
 		struct gfs_stat st2;
 
 		e = gfarm2fs_fstat(fp, &st, &st2);
 		if (e != GFARM_ERR_NO_ERROR) {
+			gfarm2fs_open_file_table_unlock();
 			gfs_stat_free(&st);
 			gfarm2fs_check_error(GFARM_MSG_2000046, OP_GETATTR,
 				"gfs_pio_stat", gfarmized.path, e);
 			free_gfarmized_path(&gfarmized);
-			gfarm2fs_open_file_table_unlock();
 			return (-gfarm_error_to_errno(e));
 		}
 		gfs_stat_free(&st);
@@ -1187,8 +1194,8 @@ gfarm2fs_utimens(const char *path, const struct timespec ts[2])
 		free_gfarmized_path(&gfarmized);
 		return (-gfarm_error_to_errno(e));
 	}
-	gfarm2fs_open_file_table_lock();
-	if ((fp = gfarm2fs_open_file_lookup(st.st_ino)) != NULL) {
+	gfarm2fs_open_file_table_rdlock();
+	if ((fp = gfarm2fs_open_file_lookup_unlocked(st.st_ino)) != NULL) {
 		open_file_wrlock(fp);
 		fp->gt[0].tv_sec = ts[0].tv_sec;
 		fp->gt[0].tv_nsec = ts[0].tv_nsec;
@@ -1217,6 +1224,8 @@ gfarm2fs_utimens(const char *path, const struct timespec ts[2])
 #define IS_WRITABLE(x) (((x) & GFARM_FILE_WRONLY) != 0 || \
 			((x) & GFARM_FILE_RDWR) != 0)
 
+static int option_unbuffered;
+
 static int
 gfs_hook_open_flags_gfarmize(int open_flags)
 {
@@ -1250,8 +1259,9 @@ gfs_hook_open_flags_gfarmize(int open_flags)
 		gfs_flags |= GFARM_FILE_EXCLUSIVE;
 #endif
 #ifdef GFARM_FILE_UNBUFFERED
-	/* open(2) and creat(2) should be unbuffered */
-	gfs_flags |= GFARM_FILE_UNBUFFERED;
+	if (option_unbuffered)
+		/* open(2) and creat(2) should be unbuffered */
+		gfs_flags |= GFARM_FILE_UNBUFFERED;
 #endif
 	return (gfs_flags);
 }
@@ -1436,7 +1446,13 @@ gfarm2fs_release(const char *path, struct fuse_file_info *fi)
 	struct gfarm2fs_file *fp = get_filep(fi);
 
 	(void) path;
-	gfarm2fs_open_file_remove(fp);
+	/*
+	 * gfarm2fs_getattr and gfarm2fs_release may be called simultaneously
+	 * after write-close.
+	 */
+	gfarm2fs_open_file_table_wrlock();
+	gfarm2fs_open_file_remove_unlocked(fp);
+
 	open_file_wrlock(fp);
 	e = gfs_pio_close(fp->gf);
 	gfarm2fs_check_error(GFARM_MSG_2000033, OP_RELEASE,
@@ -1460,6 +1476,7 @@ gfarm2fs_release(const char *path, struct fuse_file_info *fi)
 		}
 	}
 	open_file_unlock(fp);
+	gfarm2fs_open_file_table_unlock();
 	open_file_lock_destroy(fp);
 	free(fp);
 	return (-gfarm_error_to_errno(e));
@@ -1848,11 +1865,11 @@ gfarm2fs_release_cached(const char *path, struct fuse_file_info *fi)
 {
 	int rv;
 
+	rv = gfarm2fs_release(path, fi);
 	if ((fi->flags & O_ACCMODE) == O_WRONLY ||
 	    (fi->flags & O_ACCMODE) == O_RDWR ||
 	    (fi->flags & O_TRUNC) != 0)
 		uncache_path(path);
-	rv = gfarm2fs_release(path, fi);
 	gfarm2fs_replicate(path, fi);
 	return (rv);
 }
@@ -1955,6 +1972,7 @@ enum {
 	KEY_GENUINE_NLINK,
 	KEY_DISABLE_GENUINE_NLINK,
 	KEY_DIRECTORY_QUOTA_RENAME_ERROR_EXDEV,
+	KEY_UNBUFFERED,
 };
 
 #define GFARM2FS_OPT(t, p, v) \
@@ -1982,6 +2000,7 @@ static struct fuse_opt gfarm2fs_opts[] = {
 	FUSE_OPT_KEY("disable_genuine_nlink", KEY_DISABLE_GENUINE_NLINK),
 	FUSE_OPT_KEY("directory_quota_rename_error_exdev",
 	    KEY_DIRECTORY_QUOTA_RENAME_ERROR_EXDEV),
+	FUSE_OPT_KEY("unbuffered", KEY_UNBUFFERED),
 	GFARM2FS_OPT("auto_uid_min=%d", auto_uid_min, KEY_GFARM2FS_OPT),
 	GFARM2FS_OPT("auto_uid_max=%d", auto_uid_max, KEY_GFARM2FS_OPT),
 	GFARM2FS_OPT("auto_gid_min=%d", auto_gid_min, KEY_GFARM2FS_OPT),
@@ -2012,6 +2031,7 @@ usage(const char *progname, struct gfarm2fs_param *paramsp)
 "                            (default: %d)\n"
 "    -o disable_genuine_nlink use faked st_nlink\n"
 "    -o directory_quota_rename_error_exdev enable client-side directory move\n"
+"    -o unbuffered           do not use buffering in libgfarm\n"
 "    -o auto_uid_min=N       minimum UID automatically assigned (default: %d)\n"
 "    -o auto_uid_max=N       maximum UID automatically assigned (default: %d)\n"
 "    -o auto_gid_min=N       minimum GID automatically assigned (default: %d)\n"
@@ -2082,6 +2102,9 @@ gfarm2fs_opt_proc(void *data, const char *arg, int key,
 	case KEY_DIRECTORY_QUOTA_RENAME_ERROR_EXDEV:
 		paramsp->directory_quota_rename_error_exdev = 1;
 		return (0);
+	case KEY_UNBUFFERED:
+		paramsp->unbuffered = 1;
+		return (0);
 	case KEY_VERSION:
 		fprintf(stderr, "Gfarm2fs version " VERSION "\n");
 #ifdef HAVE_GFARM_VERSION
@@ -2126,6 +2149,7 @@ main(int argc, char *argv[])
 		.enable_cached_id = 0, /* for debug */
 		.genuine_nlink = 1,
 		.fix_acl = 0,
+		.unbuffered = 0,
 		.auto_uid_min = 70000,
 		.auto_uid_max = 79999,
 		.auto_gid_min = 70000,
@@ -2221,6 +2245,8 @@ main(int argc, char *argv[])
 	option_directory_quota_rename_error_exdev =
 	    params.directory_quota_rename_error_exdev;
 
+	option_unbuffered = params.unbuffered;
+
 	/* end of setting params */
 
 	gfarm2fs_replicate_init(&params);
